/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function ObmCategoriesProxy() {
  //Init logger
  this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
  var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
  this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
  this.logger.setHeader("obmCategoriesProxy: ");
  this.logger.debug("new obmCategoriesProxy()", false, false);


	this.SOAPClient = Cc["@guru4.net/soapclient;1"].getService().wrappedJSObject;

	this._listeners = [];
	this._url = null;
	this._token = {
		user: null,
		domain: null,
		session_id: null
	};
};

ObmCategoriesProxy.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICategoriesProxy) &&
			!aIID.equals(Ci.obmIProxyBase)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	/*
	 * obmIProxyBase
	 */
	get token() {
		return this._token;
	},
	set token(aToken) {
		this._token = aToken;
	},

	get WSDLUrl() {
		return this._url;
	},
	set WSDLUrl(aURL) {
		this._url = aURL;
	},

	addListener: function(listener) {
		for (var i = 0; i < this._listeners.length; ++i) {
			if (this._listeners[i] == listener)
				return;
		}
		this._listeners.push(listener);
	},

	removeListener: function(listener) {
		this._listeners = [];
	},
	_notifyListeners: function(aResultCode, aMethodName, aObj) {
		// A copy is needed, a new listeners could has been added during the callbacks
		var listeners = this._listeners.slice(0, this._listeners.length);
		for (var i = 0; i < listeners.length; ++i) {
			try {
				listeners[i].onResponse(aResultCode, aMethodName, aObj, null);
			} catch(e) {
				//Continue to notify listeners
        this.logger.error(e + " (ignored)");
			}
		}
	},

	/*
	 * obmITaskProxy
	 */
	_callback: function(aMethodName, aObj, aXml) {
    this.logger.debug("_callback(aMethodName = " + aMethodName + ")", false, false);
		if (aObj && aObj.name == "Error") {
			var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
			msg.data = aObj.fileName;
			this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, msg);
			return;
		}
		switch(aMethodName) {
			case "listCategories":
				var categories = aObj._return[0]._return;
				this._notifyListeners(Cr.NS_OK, aMethodName, new GetCategoriesShell(categories));
				break;
			default:
			this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, null);
		};
	},

	listCategories: function() {
    this.logger.debug("listCategories()", false, false);
		var p = this.SOAPClient.newSOAPClientParameters();
		// closure
		var self = this;
		this.SOAPClient.invoke(this._url, "listCategories", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml);}, null);
	}
};

function GetCategoriesShell(inner) {
	this._mInner = inner;
	this._categories = [];
	for (var i = 0; i < inner.length; i++) {
		var cat = inner[i];
		cat.QueryInterface = function(aIID) {
			if (!aIID.equals(Ci.nsISupports) &&
				!aIID.equals(Ci.obmICategory)) {
				throw Cr.NS_ERROR_NO_INTERFACE;
			}
			return this;
		};
		this._categories.push(cat);
	}
}

GetCategoriesShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmICategoriesResponse)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	getCategories: function(count) {
		count.value = this._categories.length;
		return this._categories;
	}
};


// module

var ObmCategoriesProxyModule = {
	mCID: Components.ID("{371565ce-b2ba-4dcc-bcc3-f3061be8b354}"),
	mContractID: "@obm.org/obm-sync/categoriesproxy;1",
	registerSelf: function (compMgr, fileSpec, location, type) {
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(
			this.mCID,
			"OBM Categories Proxy",
			this.mContractID,
			fileSpec,
			location,
			type
		);
	},
	getClassObject: function (compMgr, cid, iid) {
		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		return this.mFactory;
	},
	mFactory: {
		createInstance: function (outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmCategoriesProxy()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmCategoriesProxyModule;
}

