/* ***** BEGIN LICENSE BLOCK *****
 * 
 * Copyright (c) 1997-2008 Aliasource - Groupe LINAGORA
 *
 *  This program is free software; you can redistribute it and/or
 *  modify it under the terms of the GNU General Public License as
 *  published by the Free Software Foundation; either version 2 of the
 *  License, (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 * 
 *  http://www.obm.org/                                              
 * 
 * ***** END LICENSE BLOCK ***** */
const Cc = Components.classes;
const Ci = Components.interfaces;
const Cr = Components.results;

function _appendXMLElement(aRootElement, aChildElementName, aChildElementValue) {
	if ( aChildElementValue === undefined || aChildElementValue === null ) {
	  //append nothing
	} else {
	  aRootElement[aChildElementName] = aChildElementValue;
	}
}

function _xS(value) {
	// FIXME : toString() ?
	return (value.length() == 1) ? value + "" : "";
}

function _xL(value, def) {
	if (value.length() == 1)
		return parseInt(value + "", 10);
	if (def)
		return def;
	return 0;
}

function _xD(value) {
	return (value.length() == 1) ? parseFloat(value + "") : 0;
}

function _xB(value) {
	return value + "" == "true";
}

function _eventShell(aEvent, aKind) {
  aEvent.QueryInterface(Ci.obmIREventRecurrence);
  if (aKind == "event") {
		var obj = <event/>; // xmlns="http://www.obm.org/xsd/sync/event.xsd"/>;
		if (aEvent.uid) {
			obj.@id = aEvent.uid;
		}
		_appendXMLElement(obj, "extId", aEvent.extId);
  } else {
		var obj = <eventException/>
		_appendXMLElement(obj, "recurrenceId", aEvent.recurrenceId);
  }
  
  obj.@allDay = aEvent.allday;
  obj.@type = "VEVENT";
  
  _appendXMLElement(obj, "owner", aEvent.owner);
  _appendXMLElement(obj, "priority" , aEvent.priority);
  _appendXMLElement(obj, "privacy" , aEvent.privacy);
  _appendXMLElement(obj, "date" , aEvent.date);
  _appendXMLElement(obj, "duration" , aEvent.duration);
  _appendXMLElement(obj, "title" , aEvent.title);
  _appendXMLElement(obj, "description" , aEvent.description);
  _appendXMLElement(obj, "category" , aEvent.category);
  _appendXMLElement(obj, "location" , aEvent.location);
  _appendXMLElement(obj, "alert" , aEvent.alert);
  _appendXMLElement(obj, "opacity" , aEvent.opacity);

  var atts = aEvent.getAttendees({});
  for (var i = 0; i < atts.length; i++) {
    obj.attendees.attendee[i] = <attendee email={atts[i].email} displayName={atts[i].displayName} state={atts[i].status} required={atts[i].type}/>;
  }
  
  obj.recurrence = <recurrence kind={aEvent.kind} freq={aEvent.frequence} days={aEvent.days}/>;
  if (aEvent.end)
  	obj.*::recurrence.@end = aEvent.end;
  
  var exs =	aEvent.getExceptions({});
  for (var i = 0; i < exs.length; i++) {
  	obj.recurrence.exceptions.exception[i] = exs[i];
  }
  
  var eventExs = aEvent.getEventExceptions({});
  for (var i = 0; i < eventExs.length; i++) {
  	obj.recurrence.eventExceptions.eventException[i] = _eventShell(eventExs[i], "eventException");
  }
  
  return obj;
}

function JSObjEventShell(aEvent /* obmIREvent */) {
  var obj = _eventShell(aEvent, "event");
  return ("<?xml version=\"1.0\" encoding=\"UTF-8\"?>" + obj.toString());
}

function ObmRCalendarProxy() {
  // Init logger
  this.logger = Cc["@obm.org/calendar/logger;1"].createInstance(Ci.obmILogger);
  var writer = Cc["@obm.org/calendar/writer;1?type=unixfile"].getService(Ci.obmILoggerWriter);
  this.logger.setLevelWriter(this.logger.LOAD_LEVEL, writer);
  this.logger.setHeader("obmCalendarProxy: ");
  this.logger.debug("new obmCalendarProxy()", false, false);

	this.RESTClient = Cc["@obm.org/restclient;1"].getService().wrappedJSObject;

	this._listeners = [];
	this._url = null;
	this._loginUrl = null;
	this._token = {
		session_id: null
	};
};

ObmRCalendarProxy.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRCalendarProxy) &&
			!aIID.equals(Ci.obmIRProxyBase)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},

	/*
	 * obmIRProxyBase
	 */
	addListener: function(listener) {
		for (var i = 0; i < this._listeners.length; ++i) {
			if (this._listeners[i] == listener)
				return;
		}
		this._listeners.push(listener);
	},

	removeListener: function(listener) {
		this._listeners = [];
	},
	_notifyListeners: function(aResultCode, aMethodName, aObj, aData) {
		// A copy is needed, a new listeners could has been added during the
		// callbacks
		var listeners = this._listeners.slice(0, this._listeners.length);
		for (var i = 0; i < listeners.length; ++i) {
			try {
				listeners[i].onResponse(aResultCode, aMethodName, aObj, aData);
			} catch(e) {
				// Continue to notify listeners
        this.logger.error(e + " (ignored)");
			}
		}
	},

	get token() {
		return this._token;
	},
	set token(aToken) {
		this._token = aToken;
	},
	
	get PortUrl() {
		return this._url;
	},
	set PortUrl(aURL) {
		if(aURL.charAt(aURL.length -1) == '/') {
		  aURL = aURL.substring(0, aURL.length - 1);
		}
		this._url = aURL + "/calendar";
		this._loginUrl = aURL + "/login";
	},

	/*
	 * obmIRCalendarProxy
	 */
	_callback: function(aMethodName, aObj, aXml, aData) {
		this.logger.debug("_callback(aMethodName = " + aMethodName + ")", false, false);
		if (aObj && aObj.name == "Error") {
			var msg = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
			
			if (aMethodName == "doLogin" &&  aObj.fileName != "Connection failure") {
				msg.data = "No token available";
			} else {
				msg.data = aObj.fileName;
			}
		  this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, msg, aData);
		  return;
		}
		switch(aMethodName) {
			case "doLogin":
				var sid = _xS(aXml.*::sid);
				this._token.session_id = sid;
				this.logger.debug("token obtained with sid:"+sid, false, false);
				var syncVersion = null;
				if (aXml.*::version != undefined) {
				  syncVersion = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
				  syncVersion.data = aXml.*::version.@major
									  + "." + aXml.*::version.@minor
									  + "." + aXml.*::version.@release;
				  this.logger.debug("sync version:" + syncVersion.data, false, false);
				}
				this._notifyListeners(Cr.NS_OK, aMethodName, syncVersion, aData);
				break;
			case "getSync":
				this._notifyListeners(Cr.NS_OK, aMethodName, new GetSyncEventResponseShell(aXml), aData);
				break;
			case "createEvent":
				var id = Cc["@mozilla.org/supports-string;1"].createInstance(Ci.nsISupportsString);
				id.data = _xS(aXml.*::value);
				this._notifyListeners(Cr.NS_OK, aMethodName, id, aData);
				break;
			case "modifyEvent":
				this._notifyListeners(Cr.NS_OK, aMethodName, new EventShell(aXml), aData);
				break;
			case "removeEvent":
				this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
				break;
			case "removeEventByExtId":
				this._notifyListeners(Cr.NS_OK, aMethodName, null, aData);
				break;
			case "listCalendars":
				var XPArray = {
					list: aXml.*::info,
					QueryInterface: function(aIID) {
						if (!aIID.equals(Ci.nsISupports) &&
							!aIID.equals(Ci.nsIArray)) {
							throw Cr.NS_ERROR_NO_INTERFACE;
						}
						return this;
					},
					get length() {return this.list.length();},
					enumerate: function() {throw Cr.NS_ERROR_NOT_IMPLEMENTED},
					indexOf: function(idx, elt) {throw Cr.NS_ERROR_NOT_IMPLEMENTED},
					queryElementAt: function(idx, uuid) {
						var cal = new CalendarShell(this.list[idx]);
						return cal.QueryInterface(uuid);
					}
				};
				this._notifyListeners(Cr.NS_OK, aMethodName, XPArray, aData);
				break;
			case "listCategories":
				this._notifyListeners(Cr.NS_OK, aMethodName, new GetCategoriesShell(aXml));
				break;	
			default:
				this._notifyListeners(Cr.NS_ERROR_FAILURE, aMethodName, null, aData);
		};
	},

	getSync: function(aCalName, aLastSync, aData) {
    this.logger.debug("getSync()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		p.add("calendar", aCalName);
		p.add("lastSync", aLastSync);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "getSync", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	createEvent: function(aCalendarId, aEvent, aData) {
		this.logger.debug("createEvent()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		p.add("calendar", aCalendarId);
		p.add("event", JSObjEventShell(aEvent).toString());
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "createEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	modifyEvent: function(aCalendarId, aEvent, aUpdateAttendees, aData) {
    this.logger.debug("modifyEvent()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		p.add("calendar", aCalendarId);
		p.add("event", JSObjEventShell(aEvent));
		p.add("updateAttendees", aUpdateAttendees);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "modifyEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	removeEvent: function(aCalendarId, aUid, aData) {
    this.logger.debug("removeEvent()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		p.add("calendar", aCalendarId);
		p.add("id", aUid);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "removeEvent", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	removeEventByExtId: function(aCalendarId, aExtId, aData) {
    this.logger.debug("removeEventByExtId()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		p.add("calendar", aCalendarId);
		p.add("extId", aExtId);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "removeEventByExtId", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	doLogin: function(aLogin, aPwd, aOrigin, aData) {
    this.logger.debug("doLogin()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("login", aLogin);
		p.add("password", aPwd);
		p.add("origin", aOrigin);
		// closure
		var self = this;
		this.RESTClient.invoke(this._loginUrl, "doLogin", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	listCalendars: function(aData) {
    this.logger.debug("listCalendars()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "listCalendars", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml, data);}, aData);
	},
	listCategories: function(aData) {
    this.logger.debug("listCategories()", false, false);
		var p = this.RESTClient.newRESTClientParameters();
		p.add("sid", this._token.session_id);
		// closure
		var self = this;
		this.RESTClient.invoke(this._url, "listCategories", p, true, function(name, obj, xml, data) {self._callback(name, obj, xml);}, aData);
	}
};

function CalendarShell(aXml) {
	this._mXml = aXml;
}

CalendarShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRCalendarInfo)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	// obmIRCalendarInfo
	/*<info>
		<first>user3</first>
		<last>user3</last>
		<mail>user3@dom1.local</mail>
		<uid>user3</uid>
		<read>true</read>
		<write>true</write>
	</info>*/
	get uid() {return _xS(this._mXml.*::uid)},
	get firstname() {return _xS(this._mXml.*::first)},
	get lastname() {return _xS(this._mXml.*::last)},
	get mail() {return _xS(this._mXml.*::mail)},
	get read() {return _xB(this._mXml.*::read)},
	get write() {return _xB(this._mXml.*::write)}
};


function EventShell(aXml) {
	this._mXml = aXml;
}

EventShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIREventRecurrence) &&
			!aIID.equals(Ci.obmIREvent)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	// obmIREvent
	get uid() {return this._mXml.@id},
	get extId() {return this._mXml.*::extId},
	get recurrenceId() {return this._mXml.*::recurrenceId},
	get owner() {return _xS(this._mXml.*::owner)},
	get ownerEmail() {return _xS(this._mXml.*::ownerEmail)},
	get priority() {return _xL(this._mXml.*::priority)},
	get privacy() {return _xL(this._mXml.*::privacy)},
	get allday() {return _xB(this._mXml.@allDay)},
	get date() {return _xS(this._mXml.*::date)},
	get duration() {return _xL(this._mXml.*::duration)},
	get title() {return _xS(this._mXml.*::title)},
	get description() {return _xS(this._mXml.*::description)},
	get category() {return _xS(this._mXml.*::category)},
	get location() {return _xS(this._mXml.*::location)},
	get opacity() {return _xS(this._mXml.*::opacity)},
	get recurrence() {return this;},
	getAttendees: function(count) {
    var attendees = [];
    if (this._mXml.*::attendees) {
    	for each( var att in this._mXml.*::attendees.*) {
    		var attendee = {};
    		attendee.email = _xS(att.@email);
    		attendee.displayName = _xS(att.@displayName);
    		attendee.status = _xS(att.@state);
    		attendee.type = _xS(att.@required);
    		attendees.push(attendee);
    	}
    }
    //attendees = this._mXml.attendees;
		count.value = attendees.length;
		return attendees;
	},
	// obmIREventRecurrence
	get kind() {return _xS(this._mXml.*::recurrence.@kind)},
	get frequence() {return _xL(this._mXml.*::recurrence.@freq)},
	get days() {return _xS(this._mXml.*::recurrence.@days)},
	get end() {return _xS(this._mXml.*::recurrence.@end)},
	getExceptions: function(count) {
		var exceptions = [];
		var exs = this._mXml.*::recurrence.*::exceptions.*::exception.*;
		for each (var ex in exs) {
			exceptions.push(_xS(ex));
		}
		if (exceptions.length > 0) {
			count.value = exceptions.length;
			return exceptions;
		} else {
			return;
		}
	},
	getEventExceptions: function(count) {
	  var eventExceptions  = [];
	  var eventExs = this._mXml.*::recurrence.*::eventExceptions.*;
	  for each (var eventEx in eventExs) {
		  eventExceptions.push(new EventShell(eventEx));
	  }
	  if (eventExceptions.length > 0) {
		  count.value = eventExceptions.length;
		  return eventExceptions;
	  } else {
		  return;
	  }
	},
	get alert() {return _xL(this._mXml.*::alert, -1)}
};

function GetSyncEventResponseShell(aXml) {
	this._updated = [];
	this._removed = [];
	this._removedExtId = [];
	this._lastSync = aXml.@lastSync.toString(); 

	for each (var updated in aXml.*::updated.*) {
	    if ( (updated.*::recurrenceId == undefined) || (updated.*::recurrenceId == 0)) {
		  this._updated.push(new EventShell(updated));
		} else {
		  //skip this is an eventException allready in parent item
		}
	}
	
	for each (var removed in aXml.*::removed.*) {
		this._removed.push(removed.@id.toString());
		if (removed.@extId != undefined) {
			this._removedExtId.push(removed.@extId.toString());
		}
	}
	
	/*
	<calendar-changes lastSync="20081127230000000" xmlns="http://www.obm.org/xsd/sync/calendar-changes.xsd">
		<removed>
			<event id="user3-2" extId="85f148ae-2400-4860-b1dd-ae3fa062b07b"/>
			<event id="user3-3" extId="e1e253f9-b7a6-488b-a360-5ca294afb803"/>
			<event id="user3-4" extId="ad3f09e7-1a5e-40c2-ae1d-12906fbc3760"/>
			<event id="user3-5" extId="e19991ea-8131-4b61-88fe-e25b84311073"/>
		</removed>
		<updated>
			<event id="1">
			<title>eeeeee</title>
			<owner>user3</owner>
			<date>20081127110000000</date>
			<duration>3600</duration>
			<priority>2</priority>
			<privacy>0</privacy>
			<attendees>
				<attendee displayName="user3 user3" email="user3@dom1.local" required="CHAIR" state="ACCEPTED"/>
			</attendees>
			<recurrence kind="none"/>
			</event>
		</updated>
	</calendar-changes>
	*/
}

GetSyncEventResponseShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRCalendarSyncResponse)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	getUpdatedEvents: function(count) {
		count.value = this._updated.length;
		return this._updated;
	},
	getRemovedEventIds: function(count) {
		count.value = this._removed.length;
		return this._removed;
	},
	getRemovedEventExtIds: function(count) {
		count.value = this._removedExtId.length;
		return this._removedExtId;
	},
	get lastsync() {
		return this._lastSync;
	}
};

function GetCategoriesShell(aXml) {
	this._categories = [];
	for each (var xcat in aXml.*::cat) {
		var cat = {};
		cat.id = _xL(xcat.@id);
		cat.label = _xS(xcat.@label);
		cat.QueryInterface = function(aIID) {
			if (!aIID.equals(Ci.nsISupports) &&
				!aIID.equals(Ci.obmIRCategory)) {
				throw Cr.NS_ERROR_NO_INTERFACE;
			}
			return this;
		};
		this._categories.push(cat);
	}
}

GetCategoriesShell.prototype = {
	QueryInterface: function(aIID) {
		if (!aIID.equals(Ci.nsISupports) &&
			!aIID.equals(Ci.obmIRCategoriesResponse)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		return this;
	},
	getCategories: function(count) {
		count.value = this._categories.length;
		return this._categories;
	}
};
// module

var ObmRCalendarProxyModule = {
	mCID: Components.ID("{e2b34be1-b2b2-47de-bfe2-4f8126183ca5}"),
	mContractID: "@obm.org/obm-sync-rest/calendarproxy;1",
	registerSelf: function (compMgr, fileSpec, location, type) {
		dump("registerSelf:@obm.org/obm-sync-rest/calendarproxy;1\n");
		compMgr = compMgr.QueryInterface(Ci.nsIComponentRegistrar);
		compMgr.registerFactoryLocation(
			this.mCID,
			"OBM Calendar Proxy",
			this.mContractID,
			fileSpec,
			location,
			type
		);
	},
	getClassObject: function (compMgr, cid, iid) {
		if (!cid.equals(this.mCID)) {
			throw Cr.NS_ERROR_NO_INTERFACE;
		}
		if (!iid.equals(Ci.nsIFactory)) {
			throw Cr.NS_ERROR_NOT_IMPLEMENTED;
		}
		return this.mFactory;
	},
	mFactory: {
		createInstance: function (outer, iid) {
			if (outer != null) {
				throw Cr.NS_ERROR_NO_AGGREGATION;
			}
			return (new ObmRCalendarProxy()).QueryInterface(iid);
		}
	},

	canUnload: function(compMgr) {
		return true;
	}
};

function NSGetModule(compMgr, fileSpec) {
	return ObmRCalendarProxyModule;
}

